// -*- mode: c++ -*-
// Copyright 2016-2022 Keyboardio, inc. <jesse@keyboard.io>
// See "LICENSE" for license details

#ifndef BUILD_INFORMATION
#define BUILD_INFORMATION "locally built on " __DATE__ " at " __TIME__
#endif

#include "Kaleidoscope.h"

#include "Kaleidoscope-EEPROM-Settings.h"
#include "Kaleidoscope-EEPROM-Keymap.h"

#include "Kaleidoscope-FocusSerial.h"

#include "Kaleidoscope-MouseKeys.h"

#include "Kaleidoscope-Macros.h"

#include "Kaleidoscope-LEDControl.h"

#include "Kaleidoscope-LEDEffect-BootGreeting.h"

#include "Kaleidoscope-LEDEffect-Rainbow.h"

#include "Kaleidoscope-LED-Stalker.h"

#include "Kaleidoscope-LED-Palette-Theme.h"

#include "Kaleidoscope-Colormap.h"

#include "Kaleidoscope-HardwareTestMode.h"

#include "Kaleidoscope-HostPowerManagement.h"

#include "Kaleidoscope-MagicCombo.h"

#include "Kaleidoscope-USB-Quirks.h"

#include "Kaleidoscope-OneShot.h"
#include "Kaleidoscope-Escape-OneShot.h"

enum {
  MACRO_VERSION_INFO,
  MACRO_ANY,
};

/** The Model 100's key layouts are defined as 'keymaps'.
 *
 * Each keymap is defined as a list using the 'KEYMAP_STACKED' macro, built
 * of first the left hand's layout, followed by the right hand's layout.
 *
 * Keymaps typically consist mostly of `Key_` definitions. There are many, many
 * keys defined as part of the USB HID Keyboard specification. You can find the
 * names (if not yet the explanations) for all the standard `Key_` defintions
 * offered by Kaleidoscope in these files:
 *    https://github.com/keyboardio/Kaleidoscope/blob/master/src/kaleidoscope/key_defs/keyboard.h
 *    https://github.com/keyboardio/Kaleidoscope/blob/master/src/kaleidoscope/key_defs/consumerctl.h
 *    https://github.com/keyboardio/Kaleidoscope/blob/master/src/kaleidoscope/key_defs/sysctl.h
 *    https://github.com/keyboardio/Kaleidoscope/blob/master/src/kaleidoscope/key_defs/keymaps.h
 *
 * Additional things that should be documented here include
 *   using ___ to let keypresses fall through to the previously active layer
 *   using XXX to mark a keyswitch as 'blocked' on this layer
 *   using ShiftToLayer() and LockLayer() keys to change the active keymap.
 *   keeping NUM and FN consistent and accessible on all layers
 *
 * A key defined as 'ShiftToLayer(FUNCTION)' will switch to FUNCTION while held.
 * Similarly, a key defined as 'LockLayer(FUNCTION)' will switch to FUNCTION
 * when tapped.
 */

enum {
  PRIMARY,
  SHIFTED,
  ALTGR,
  ALTGR_SHIFT,
  FUNCTION,
  HUN,
  COLEMAK,
  ONEHAND_LEFT,
  ONEHAND_RIGHT,
};

/* This comment temporarily turns off astyle's indent enforcement
 *   so we can make the keymaps actually resemble the physical key layout better
 */
// clang-format off

KEYMAPS(
  [PRIMARY] = KEYMAP_STACKED
  (___,                     Key_1,           Key_2,     Key_3,      Key_4, Key_5, Key_LEDEffectNext,
   RALT(Key_7),             LSHIFT(Key_6),   Key_Comma, Key_Period, Key_P, Key_Z, Key_Tab,
   ___,                     Key_A,           Key_O,     Key_E,      Key_U, Key_I,
   LALT(LCTRL(Key_Delete)), RALT(Key_Comma), Key_Q,     Key_J,      Key_K, Key_X, Key_Escape,
   Key_LeftControl, Key_Backspace, Key_LeftAlt, OSL(SHIFTED),
   OSL(FUNCTION),

   M(1),        Key_6, Key_7, Key_8, Key_9, Key_Backtick, MoveToLayer(HUN),
   Key_Enter,   Key_F, Key_G, Key_C, Key_R, Key_L,        LSHIFT(Key_7),
                Key_D, Key_H, Key_T, Key_N, Key_S,        LSHIFT(Key_1),
   Key_LeftGui, Key_B, Key_M, Key_W, Key_V, Key_Y,        Key_Slash,
   OSL(SHIFTED), OSL(ALTGR), Key_Spacebar, Key_RightControl,
   OSL(FUNCTION)),

  [SHIFTED] = KEYMAP_STACKED
  (___,         LSHIFT(Key_4),      RALT(Key_V),   RALT(Key_X),      RALT(Key_Semicolon), LSHIFT(Key_5), ___,
   RALT(Key_1), LSHIFT(Key_Comma),  RALT(Key_M),   RALT(Key_Period), LSHIFT(Key_P),       LSHIFT(Key_Z), LSHIFT(Key_Tab),
   ___,         LSHIFT(Key_A),      LSHIFT(Key_O), LSHIFT(Key_E),    LSHIFT(Key_U),       LSHIFT(Key_I),
   ___,         LSHIFT(Key_Period), LSHIFT(Key_Q), LSHIFT(Key_J),    LSHIFT(Key_K),       LSHIFT(Key_X), ___,
   ___, LSHIFT(Key_Backspace), ___, ___,
   ___,

   M(0),              RALT(Key_3),   RALT(Key_C),   RALT(Key_Slash), LSHIFT(Key_8), LSHIFT(Key_9), MoveToLayer(COLEMAK),
   LSHIFT(Key_Enter), LSHIFT(Key_F), LSHIFT(Key_G), LSHIFT(Key_C),   LSHIFT(Key_R), LSHIFT(Key_L), LSHIFT(Key_3),
                      LSHIFT(Key_D), LSHIFT(Key_H), LSHIFT(Key_T),   LSHIFT(Key_N), LSHIFT(Key_S), LSHIFT(Key_2),
   ___,               LSHIFT(Key_B), LSHIFT(Key_M), LSHIFT(Key_W),   LSHIFT(Key_V), LSHIFT(Key_Y), LSHIFT(Key_Slash),
   ___, OSL(ALTGR_SHIFT), ___, ___,
   ___),

  [ALTGR] = KEYMAP_STACKED
  (___,                XXX,             RALT(Key_2),     RALT(LSHIFT(Key_3)),        RALT(Key_U),                    RALT(Key_5),                     ___,
   RALT(Key_Backtick), RALT(Key_A),     Key_0,           RALT(LSHIFT(Key_Period)),   Key_Minus,                      RALT(Key_Z),                     ___,
   ___,                Key_Quote,       Key_Equals,      Key_Semicolon,              Key_RightBracket,               Key_NonUsBackslashAndPipe,
   ___,                RALT(Key_Quote), Key_LeftBracket, RALT(Key_J),                Key_Backslash,                  RALT(Key_NonUsBackslashAndPipe), ___,
   ___, ___, ___, OSL(ALTGR_SHIFT),
   ___,

   ___, XXX,                 XXX,         RALT(LSHIFT(Key_Slash)), RALT(Key_8),         RALT(LSHIFT(Key_8)),     MoveToLayer(ONEHAND_LEFT),
   ___, RALT(Key_I),         RALT(Key_L), RALT(Key_E),             RALT(Key_H),         RALT(Key_S),             RALT(Key_RightBracket),
        RALT(Key_D),         XXX,         RALT(Key_T),             RALT(Key_K),         RALT(Key_Minus),         RALT(Key_Equals),
   ___, RALT(LSHIFT(Key_O)), RALT(Key_O), RALT(Key_P),             RALT(LSHIFT(Key_M)), RALT(LSHIFT(Key_Comma)), RALT(Key_Y),
   OSL(ALTGR_SHIFT), ___, ___, ___,
   ___),

  [ALTGR_SHIFT] = KEYMAP_STACKED
  (___,         XXX,                      RALT(LSHIFT(Key_2)),     XXX,                   RALT(Key_Backslash),      RALT(LSHIFT(Key_5)),                     ___,
   RALT(Key_R), RALT(LSHIFT(Key_A)),      LSHIFT(Key_0),           XXX,                   LSHIFT(Key_Minus),        RALT(LSHIFT(Key_Z)),                     ___,
   ___,         LSHIFT(Key_Quote),        LSHIFT(Key_Equals),      LSHIFT(Key_Semicolon), LSHIFT(Key_RightBracket), LSHIFT(Key_NonUsBackslashAndPipe),
   ___,         RALT(LSHIFT(Key_Quote)),  LSHIFT(Key_LeftBracket), RALT(LSHIFT(Key_J)),   LSHIFT(Key_Backslash),    RALT(LSHIFT(Key_NonUsBackslashAndPipe)), ___,
   ___, ___, ___, ___,
   ___,

   ___, RALT(LSHIFT(Key_C)), RALT(LSHIFT(Key_R)), RALT(LSHIFT(Key_V)), RALT(LSHIFT(Key_Semicolon)), XXX,                     ___,
   ___, RALT(LSHIFT(Key_I)), RALT(LSHIFT(Key_L)), RALT(LSHIFT(Key_E)), RALT(LSHIFT(Key_H)),         RALT(LSHIFT(Key_S)),     RALT(LSHIFT(Key_4)),
        RALT(LSHIFT(Key_D)), XXX,                 RALT(LSHIFT(Key_T)), RALT(LSHIFT(Key_K)),         RALT(LSHIFT(Key_Minus)), RALT(LSHIFT(Key_Equals)),
   ___, RALT(LSHIFT(Key_7)), RALT(LSHIFT(Key_B)), RALT(LSHIFT(Key_N)), RALT(LSHIFT(Key_0)),         RALT(Key_0),             RALT(LSHIFT(Key_Y)),
   ___, ___, ___, ___,
   ___),

  [FUNCTION] = KEYMAP_STACKED
  (Key_PrintScreen, Key_F1,          Key_F2,      Key_F3,        Key_F4,        Key_F5,            Key_CapsLock,
   Key_Tab,         ___,             Key_mouseUp, Key_mouseBtnN, Key_mouseBtnR, Key_mouseScrollDn, Key_PageUp,
   Key_Home,        Key_mouseL,      Key_mouseDn, Key_mouseR,    Key_mouseBtnL, Key_mouseScrollUp,
   Key_End,         Key_PrintScreen, Key_Insert,  Key_mouseBtnP, Key_mouseBtnM, ___,               Key_PageDown,
   ___, Key_Delete, ___, Key_LeftShift,
   ___,

   Consumer_ScanPreviousTrack, Key_F6,                 Key_F7,                   Key_F8,                   Key_F9,         Key_F10,     Key_F11,
   Consumer_PlaySlashPause,    Consumer_ScanNextTrack, RALT(Key_B),              RALT(Key_N),              RALT(Key_F),    RALT(Key_G), Key_F12,
   ___,                        Key_LeftArrow,          Key_DownArrow,            Key_UpArrow,              Key_RightArrow, RALT(Key_X),
   Key_PcApplication,          Consumer_Mute,          Consumer_VolumeDecrement, Consumer_VolumeIncrement, ___,            RALT(Key_Q), RALT(Key_W),
   Key_RightShift, Key_RightAlt, Key_Enter, ___,
   ___),

  [HUN] = KEYMAP_STACKED
  (___,          Key_1, Key_2, Key_3, Key_4, Key_5, Key_LEDEffectNext,
   Key_Backtick, Key_Q, Key_W, Key_E, Key_R, Key_T, Key_Tab,
   ___,          Key_A, Key_S, Key_D, Key_F, Key_G,
   ___,          Key_Z, Key_X, Key_C, Key_V, Key_B, Key_Escape,
   Key_LeftControl, Key_Backspace, Key_LeftAlt, Key_LeftShift,
   OSL(FUNCTION),

   ___,         Key_6, Key_7, Key_8,     Key_9,      Key_0,         MoveToLayer(PRIMARY),
   Key_Enter,   Key_Y, Key_U, Key_I,     Key_O,      Key_P,         Key_Minus,
                Key_H, Key_J, Key_K,     Key_L,      Key_Semicolon, Key_Quote,
   Key_LeftGui, Key_N, Key_M, Key_Comma, Key_Period, Key_Slash,     Key_Backslash,
   Key_RightShift, Key_RightAlt, Key_Spacebar, Key_RightControl,
   OSL(FUNCTION)),

  [COLEMAK] = KEYMAP_STACKED
  (___,                       Key_1, Key_2, Key_3, Key_4, Key_5, ___,
   Key_Backtick,              Key_Q, Key_W, Key_F, Key_P, Key_G, Key_Tab,
   ___,                       Key_A, Key_R, Key_S, Key_T, Key_D,
   Key_NonUsBackslashAndPipe, Key_Y, Key_X, Key_C, Key_V, Key_B, Key_Escape,
   Key_LeftControl, Key_Backspace, Key_LeftAlt, Key_LeftShift,
   OSL(FUNCTION),

   ___,         Key_6, Key_7, Key_8,     Key_9,      Key_0,         MoveToLayer(PRIMARY),
   Key_Enter,   Key_J, Key_L, Key_U,     Key_Z,      Key_Semicolon, Key_Equals,
                Key_H, Key_N, Key_E,     Key_I,      Key_O,         Key_Quote,
   Key_LeftGui, Key_K, Key_M, Key_Comma, Key_Period, Key_Slash,     Key_Minus,
   Key_RightShift, Key_RightAlt, Key_Spacebar, Key_RightControl,
   OSL(FUNCTION)),

  [ONEHAND_LEFT] = KEYMAP_STACKED
  (XXX, Key_1, Key_2, Key_3, Key_4, Key_5, Key_Tab,
   XXX, Key_Q, Key_W, Key_E, Key_R, Key_T, Key_Enter,
   XXX, Key_A, Key_S, Key_D, Key_F, Key_G,
   XXX, Key_Z, Key_X, Key_C, Key_V, Key_B, Key_Escape,
   Key_LeftControl, Key_Spacebar, Key_LeftAlt, Key_LeftShift,
   OSL(ONEHAND_RIGHT),

   XXX, XXX, XXX, XXX, XXX, XXX, XXX,
   XXX, XXX, XXX, XXX, XXX, XXX, XXX,
        XXX, XXX, XXX, XXX, XXX, XXX,
   ___, XXX, XXX, XXX, XXX, XXX, XXX,
   XXX, XXX, XXX, XXX,
   XXX),

  [ONEHAND_RIGHT] = KEYMAP_STACKED
  (___, Key_6, Key_7, Key_8,     Key_9,      Key_Backtick,  MoveToLayer(PRIMARY),
   ___, Key_Y, Key_U, Key_I,     Key_O,      Key_P,         ___,
   ___, Key_H, Key_J, Key_K,     Key_L,      Key_Semicolon,
   ___, Key_N, Key_M, Key_Comma, Key_Period, Key_Slash,     ___,
   Key_LeftControl, Key_Backspace, Key_LeftAlt, Key_LeftShift,
   ___,

   XXX, XXX, XXX, XXX, XXX, XXX, XXX,
   XXX, XXX, XXX, XXX, XXX, XXX, XXX,
        XXX, XXX, XXX, XXX, XXX, XXX,
   ___, XXX, XXX, XXX, XXX, XXX, XXX,
   XXX, XXX, XXX, XXX,
   XXX)
) // KEYMAPS(

/* Re-enable astyle's indent enforcement */
// clang-format on

static void versionInfoMacro(uint8_t key_state) {
  if (keyToggledOn(key_state)) {
    Macros.type(PSTR("Keyboardio Model 100 - Kaleidoscope "));
    Macros.type(PSTR(BUILD_INFORMATION));
  }
}

static void anyKeyMacro(KeyEvent &event) {
  if (keyToggledOn(event.state)) {
    event.key.setKeyCode(Key_A.getKeyCode() + (uint8_t)(millis() % 36));
    event.key.setFlags(0);
  }
}

const macro_t *macroAction(uint8_t macro_id, KeyEvent &event) {
  switch (macro_id) {

  case MACRO_VERSION_INFO:
    versionInfoMacro(event.state);
    break;

  case MACRO_ANY:
    anyKeyMacro(event);
    break;
  }
  return MACRO_NONE;
}

void toggleLedsOnSuspendResume(
    kaleidoscope::plugin::HostPowerManagement::Event event) {
  switch (event) {
  case kaleidoscope::plugin::HostPowerManagement::Suspend:
    LEDControl.disable();
    break;
  case kaleidoscope::plugin::HostPowerManagement::Resume:
    LEDControl.enable();
    break;
  case kaleidoscope::plugin::HostPowerManagement::Sleep:
    break;
  }
}

void hostPowerManagementEventHandler(
    kaleidoscope::plugin::HostPowerManagement::Event event) {
  toggleLedsOnSuspendResume(event);
}

enum {
  COMBO_TOGGLE_NKRO_MODE,
  COMBO_ENTER_TEST_MODE
};

static void toggleKeyboardProtocol(uint8_t combo_index) {
  USBQuirks.toggleKeyboardProtocol();
}

static void enterHardwareTestMode(uint8_t combo_index) {
  HardwareTestMode.runTests();
}

USE_MAGIC_COMBOS({.action = toggleKeyboardProtocol,
                  // Left Fn + Esc + Shift
                  .keys = {R3C6, R2C6, R3C7}},
                 {.action = enterHardwareTestMode,
                  // Left Fn + Prog + LED
                  .keys = {R3C6, R0C0, R0C6}});

KALEIDOSCOPE_INIT_PLUGINS(
  EEPROMSettings,
  EEPROMKeymap,
  Focus,
  FocusSettingsCommand,
  FocusEEPROMCommand,
  BootGreetingEffect,
  HardwareTestMode,
  LEDControl,
  LEDOff,
  LEDRainbowWaveEffect,
  StalkerEffect,
  LEDPaletteTheme,
  ColormapEffect,
  Macros,
  MouseKeys,
  HostPowerManagement,
  MagicCombo,
  USBQuirks,
  OneShot,
  EscapeOneShot,
  EscapeOneShotConfig
);

void setup() {
  Kaleidoscope.setup();
  LEDRainbowWaveEffect.brightness(255);
  HardwareTestMode.setActionKey(R3C6);
  StalkerEffect.variant = STALKER(BlazingTrail);
  StalkerEffect.activate();
  EEPROMKeymap.setup(9);
  ColormapEffect.max_layers(9);
  MouseKeys.speed = 2;
  MouseKeys.setSpeedLimit(31);
  MouseKeys.wheelDelay = 60;
}

void loop() {
  Kaleidoscope.loop();
}
